import os
import pandas as pd


class CustomAggregation:

    def __init__(self, aggr_admin, misc_files_dir, province_column, adjustment=False):
        self.admin_stats = {}
        self.admin_classification = {}
        self.admin_province_mapping = {}
        self.admin_to_hr = {}

        self.adj_dict = {}
        self.hr_average = {}
        self.month_average = {}
        self.rural_urban_average = {}
        self.hr_month_average = {}
        self.hr_rural_urban_average = {}
        self.rural_urban_month_average = {}

        self.aggr_admin = aggr_admin
        self.misc_files_dir = misc_files_dir
        self.province_column = province_column
        self.create_admin_mappings()
        self.adjustment = adjustment

        self.create_admin_mappings()
        self.set_adjustment_numbers()

    def create_admin_mappings(self):

        self.admin_classification[self.aggr_admin] = pd.read_csv(
            '{}/admin4_classification.csv'.format(self.misc_files_dir))
        self.admin_classification[self.aggr_admin] = self.admin_classification[self.aggr_admin][
            ~self.admin_classification[self.aggr_admin]['classification'].isnull()]
        self.admin_classification[self.aggr_admin][self.aggr_admin] = self.admin_classification[self.aggr_admin][
            self.aggr_admin].astype(str).str.zfill(6)

        self.admin_province_mapping[self.aggr_admin] = pd.read_csv(
            '{}/admin4_province_mapping.csv'.format(self.misc_files_dir))
        self.admin_province_mapping[self.aggr_admin]['adm4_code'] = self.admin_province_mapping[self.aggr_admin][
            self.aggr_admin].astype(str).str.zfill(6)

        telangana_admin4s = pd.read_csv('{}/telangana_admin4s.csv'.format(self.misc_files_dir), header=None)
        telangana_admin4s[0] = telangana_admin4s[0].astype(str).str.zfill(6)

        self.admin_province_mapping[self.aggr_admin].loc[
            self.admin_province_mapping[self.aggr_admin]['adm4_code'].isin(telangana_admin4s[0]), 'adm1_code'] = None
        self.admin_province_mapping[self.aggr_admin].loc[
            self.admin_province_mapping[self.aggr_admin]['adm4_code'].isin(
                telangana_admin4s[0]), 'adm1_name'] = 'Telangana'

        # cellphone adjustment
        self.admin_to_hr[self.aggr_admin] = pd.read_csv('{}/admin4_to_hr.csv'.format(self.misc_files_dir))
        self.admin_to_hr[self.aggr_admin][self.aggr_admin.lower()] = self.admin_to_hr[self.aggr_admin][
            self.aggr_admin.lower()
        ].astype(str).str.zfill(6)

    def set_adjustment_numbers(self):
        hr_cellphone_coverage = pd.read_excel('{}/Results.xlsx'.format(self.misc_files_dir),
                                              sheet_name='% people with mobiles')

        # extract 2020 cellphone adjustment numbers and prepare the dictionary
        hrcc_2020 = hr_cellphone_coverage[hr_cellphone_coverage['month'].str.contains('2020')]
        hrcc_2020['month'] = hrcc_2020['month'].str[-1]

        self.adj_dict = dict(
            (x, dict((x_t, dict(zip(y_t['month'].str[-1], y_t['% people with mobiles']))) for (x_t, y_t) in
                     y.groupby('Unnamed: 3'))) for x, y in hrcc_2020.groupby('hr2'))

        # each feature average
        self.rural_urban_average = dict(hrcc_2020.groupby('Unnamed: 3')['% people with mobiles'].mean())
        self.month_average = dict(hrcc_2020.groupby('month')['% people with mobiles'].mean())
        self.hr_average = dict(hrcc_2020.groupby('hr2')['% people with mobiles'].mean())

        # two features combined average
        self.rural_urban_month_average = dict(
            hrcc_2020.groupby(['Unnamed: 3', 'month'])['% people with mobiles'].mean())
        self.hr_month_average = dict(hrcc_2020.groupby(['hr2', 'month'])['% people with mobiles'].mean())
        self.hr_rural_urban_average = dict(hrcc_2020.groupby(['hr2', 'Unnamed: 3'])['% people with mobiles'].mean())

    def read_admin_stats(self, spark_export_location):

        # read the CSV file saved
        admin_stats_csv_file = [i for i in os.listdir(spark_export_location) if
                                i.startswith('part') and i.endswith('.csv')][0]

        self.admin_stats[self.aggr_admin] = pd.read_csv(spark_export_location + admin_stats_csv_file)
        self.admin_stats[self.aggr_admin]['adm4_code_home'] = self.admin_stats[self.aggr_admin][
            'adm4_code_home'].astype(str).str.zfill(6)

    def classify(self):
        # add rural/urban information for admin4s
        self.admin_stats[self.aggr_admin][self.aggr_admin + '_home'] = self.admin_stats[self.aggr_admin][
            self.aggr_admin + '_home'].astype(str).str.zfill(6)
        self.admin_stats[self.aggr_admin]['classification'] = self.admin_stats[self.aggr_admin][
            self.aggr_admin + '_home'].map(
            dict(zip(self.admin_classification[self.aggr_admin][self.aggr_admin],
                     self.admin_classification[self.aggr_admin]['classification']))
        )

    def _adjust(self, row, month):
        """
        `adjust` function to adjust the counts as per mobile coverage percentage.
        """
        if str(row['hr']) == 'nan':
            adjustment_factor = self.rural_urban_month_average[(row['classification'].lower(), str(int(month)))]
        elif row['classification'].lower() not in self.adj_dict[row['hr']]:
            if (int(row['hr']), row['classification'].lower()) not in self.hr_rural_urban_average:
                adjustment_factor = self.hr_month_average[(int(row['hr']), str(int(month)))]
            else:
                adjustment_factor = self.hr_rural_urban_average[(int(row['hr']), row['classification'].lower())]
        elif str(int(month)) not in self.adj_dict[row['hr']][row['classification'].lower()]:
            near_month = str(
                min(map(int, self.adj_dict[row['hr']][row['classification'].lower()]),
                    key=lambda x: abs(x - int(month))))
            adjustment_factor = self.adj_dict[row['hr']][row['classification'].lower()][near_month]
        elif self.adj_dict[row['hr']][row['classification'].lower()][str(int(month))] == 0:
            adjustment_factor = 0.1
        else:
            adjustment_factor = self.adj_dict[row['hr']][row['classification'].lower()][str(int(month))]

        return row['total_samples'] / adjustment_factor, row['no_of_stms'] / adjustment_factor

    def adjust(self):
        # adjustment code
        self.admin_stats[self.aggr_admin]['hr'] = self.admin_stats[self.aggr_admin][self.aggr_admin + '_home'].map(
            dict(zip(self.admin_to_hr[self.aggr_admin][self.aggr_admin.lower()],
                     self.admin_to_hr[self.aggr_admin]['hr'])))
        self.admin_stats[self.aggr_admin]['adjusted_stats'] = self.admin_stats[self.aggr_admin].apply(
            self._adjust, axis=1, args=(self.month,)
        )
        self.admin_stats[self.aggr_admin]['adjusted_total_samples'] = \
            self.admin_stats[self.aggr_admin]['adjusted_stats'].str[0]
        self.admin_stats[self.aggr_admin]['adjusted_no_of_stms'] = \
            self.admin_stats[self.aggr_admin]['adjusted_stats'].str[1]

    def export_province_wise_stm(self):

        self.admin_stats['province'] = self.admin_stats[self.aggr_admin][self.aggr_admin + '_home'].map(
            dict(zip(self.admin_province_mapping[self.aggr_admin][self.aggr_admin],
                     self.admin_province_mapping[self.aggr_admin][self.province_column])))
        self.admin_stats[self.aggr_admin]['province'] = self.admin_stats[self.aggr_admin][
            self.aggr_admin + '_home'].map(
            dict(zip(self.admin_province_mapping[self.aggr_admin][self.aggr_admin],
                     self.admin_province_mapping[self.aggr_admin][self.province_column])))

        province_rural_urban_stm = self.admin_stats[self.aggr_admin].groupby(
            ['classification', 'province']).sum().reset_index()

        province_rural_urban_stm['percentage'] = province_rural_urban_stm['no_of_stms'] / province_rural_urban_stm[
            'total_samples']

        if self.adjustment:
            province_rural_urban_stm['adjusted_percentage'] = province_rural_urban_stm['adjusted_no_of_stms'] / \
                                                              province_rural_urban_stm['adjusted_total_samples']

        return province_rural_urban_stm.sort_values(by=['percentage'], ascending=False)

    def aggregate(self, spark_export_location):
        self.read_admin_stats(spark_export_location)

        self.classify()

        if self.adjustment:
            self.adjust()

        return self.export_province_wise_stm()



import pyspark.sql.functions as F


class MobilitySummary:

    def __init__(self, delta_loc, adm_col, quality_filter=5, spark=None):
        self.sjr = spark.read.format('delta').option("header", "true").load(delta_loc)
        self.adm_col = adm_col

        if quality_filter > 0:
            self.sjr = self.quality_control_filter(self.sjr, quality_filter)

    @staticmethod
    def quality_control_filter(sdf, records):
        # filter out the IDs which have less than "records" number of records
        d = sdf.groupby('device_id').count().where(F.col('count') >= records)
        sdf = sdf.join(d, on=["device_id"], how="inner")
        return sdf

    def _get_records_per_admin(self, adm_col):
        mobility_record_count = self.sjr.groupBy(adm_col).count()
        mobility_record_count = mobility_record_count.withColumnRenamed("count", "frequency")
        return mobility_record_count

    def _get_devices_per_admin(self, adm_col):
        mobility_device_count = self.sjr.groupBy(adm_col).agg(F.countDistinct("device_id")) \
            .withColumnRenamed("count(DISTINCT device_id)", "unique_devices")
        return mobility_device_count

    def _get_records_per_date_admin(self, adm_col):
        sjr_with_date = self.sjr.withColumn("date", F.from_unixtime("timestamp", format="yyyy_MM_dd"))
        mobility_record_date_count = sjr_with_date.groupBy(adm_col, 'date').count().withColumnRenamed("count", "num_records")
        return mobility_record_date_count

    def _get_devices_per_date_admin(self, adm_col):
        sjr_with_date = self.sjr.withColumn("date", F.from_unixtime("timestamp", format="yyyy_MM_dd"))
        mobility_device_date_count = sjr_with_date.groupBy(adm_col, 'date').agg(
            F.countDistinct("device_id")
        ).withColumnRenamed("count(DISTINCT device_id)", "unique_devices")
        return mobility_device_date_count

    def get_records_per_admin(self):
        return self._get_records_per_admin(self.adm_col)

    def get_records_per_date_admin(self):
        return self._get_records_per_date_admin(self.adm_col)

    def get_devices_per_admin(self):
        return self._get_devices_per_admin(self.adm_col)

    def get_devices_per_date_admin(self):
        return self._get_devices_per_date_admin(self.adm_col)

    def get_frequency_map(self):
        frequency_table = self.sjr.groupBy('device_id') \
            .count() \
            .withColumnRenamed("count", "frequency") \
            .groupBy('frequency') \
            .count() \
            .orderBy(F.asc("frequency"))
        return frequency_table



import time
from datetime import datetime
from dateutil import relativedelta
import pandas as pd

from pyspark.sql import functions as F
from pyspark.sql.functions import from_unixtime, col, lit, month as mth
import pyspark.sql.types as T


class GenericShortTermMigration:

    def __init__(self, spark_session, aggr_admin, home_admin, raw_file_location, stm_range, raw_file_format='delta',
                 distance_filter=False, quality_control_samples=5, home_duration=12, distance=0, end_at_home=False,
                 has_month_filter=False, at_a_stretch=True, fixed_home_period=False,
                 home_period=('2019/01/01', '2020/01/01'),
                 left_padding_days=15, right_padding_days=15, timeframe_months=1):

        spark = spark_session
        self.quality_control_samples = quality_control_samples
        self.aggr_admin = aggr_admin
        self.home_admin = home_admin
        self.sjr_master = spark.read.format(raw_file_format).option("header", "true").load(raw_file_location)
        self.sjr = None
        self.stm_range = stm_range
        self.distance_filter = distance_filter
        self.distance = 0
        self.end_at_home = end_at_home
        self.has_month_filter = has_month_filter
        self.at_a_stretch = at_a_stretch
        self.left_padding_days = left_padding_days
        self.right_padding_days = right_padding_days
        self.timeframe_months = timeframe_months
        self.fixed_home_period = fixed_home_period

        if distance_filter:
            self.set_dist_map(distance)

        if fixed_home_period:
            self.set_device_id_homes(home_period, home_duration)

    def set_device_id_homes(self, home_period, home_duration):

        fixed_period_sdf = self.extract_data(self.sjr_master, home_period[0], home_duration)
        fixed_period_sdf = self.quality_control_filter(fixed_period_sdf, self.quality_control_samples)

        # convert UNIX timestamp to date format
        fixed_period_sdf = fixed_period_sdf.withColumn('date', from_unixtime('timestamp').cast(T.DateType()))

        date_device_sdf = self.date_device_mode(fixed_period_sdf)
        self.device_id_homes = self.append_mode_admin(date_device_sdf, self.aggr_admin)

        if self.aggr_admin != self.home_admin:
            self.device_id_homes = self.append_mode_admin(self.device_id_homes, self.home_admin)

        if self.aggr_admin == self.home_admin:
            self.device_id_homes = self.device_id_homes.dropDuplicates(['device_id']).select(
                col('device_id'), col(self.aggr_admin + '_home')
            )
        else:
            self.device_id_homes = self.device_id_homes.dropDuplicates(['device_id']).select(
                col('device_id'), col(self.aggr_admin + '_home'), col(self.home_admin + '_home')
            )

        # if we want to save home locations for later
        # self.device_id_homes.write.format("delta").mode("overwrite").save("/mnt/CUBEIQ/esapv/India/Unacast/Short Term Migration/One Month Definition/Device ID Homes")

    @staticmethod
    def extract_data(sdf, start_date, no_of_months, by='timestamp'):
        # get the from and to value for the interval
        from_t = datetime.strptime(start_date, "%Y/%m/%d")
        to_t = from_t + relativedelta.relativedelta(months=no_of_months)

        if by == 'timestamp':
            from_t = time.mktime(from_t.timetuple())
            to_t = time.mktime(to_t.timetuple())

        # filter by date range
        timestamps = (from_t, to_t)
        print(timestamps)

        return sdf.where(col(by).between(*timestamps))

    @staticmethod
    def extract_data_by_ends(sdf, start_date, end_date, by='timestamp'):
        # get the from and to value for the interval
        from_t = datetime.strptime(start_date, "%Y/%m/%d")
        to_t = datetime.strptime(end_date, "%Y/%m/%d")

        if by == 'timestamp':
            from_t = time.mktime(from_t.timetuple())
            to_t = time.mktime(to_t.timetuple())

        # filter by date range
        timestamps = (from_t, to_t)
        print(timestamps)

        return sdf.where(col(by).between(*timestamps))

    @staticmethod
    def quality_control_filter(sdf, records):
        # filter out the IDs which have less than "records" number of records
        d = sdf.groupby('device_id').count().where(col('count') >= records)
        sdf = sdf.join(d, on=["device_id"], how="inner")
        return sdf

    def date_device_mode(self, sdf):
        # select MODE L4_CODE for unique device_id and date, so that each device has a L4_CODE for each date

        if self.aggr_admin == self.home_admin:
            device_date_admin_counts = sdf.groupBy(
                ['device_id', 'date', self.aggr_admin]
            ).count()

            sdf = (device_date_admin_counts.groupBy('device_id', 'date').agg(
                F.max(
                    F.struct(F.col('count'), F.col(self.aggr_admin))
                ).alias('max')).select(F.col('device_id'),
                                       F.col('date'),
                                       F.col('max.{}'.format(self.aggr_admin))
                                       )
                   )
            return sdf
        else:
            device_date_admin_counts = sdf.groupBy(
                ['device_id', 'date', self.home_admin, self.aggr_admin]
            ).count()

            sdf = (device_date_admin_counts.groupBy('device_id', 'date').agg(
                F.max(
                    F.struct(F.col('count'), F.col(self.aggr_admin), F.col(self.home_admin))
                ).alias('max')).select(F.col('device_id'),
                                       F.col('date'),
                                       F.col('max.{}'.format(self.home_admin)),
                                       F.col('max.{}'.format(self.aggr_admin))
                                       )
                   )
            return sdf

    @staticmethod
    def append_mode_admin(sdf, column_name):
        # find home by finding mode of L4_CODE
        device_admin_counts = sdf.groupBy(['device_id', column_name]).count()
        result = device_admin_counts.groupBy('device_id').agg(
            F.max(
                F.struct(F.col('count'), F.col(column_name))
            ).alias('max')).select(F.col('device_id'), F.col('max.{}'.format(column_name)))

        sdf = sdf.join(
            result.withColumnRenamed(column_name, column_name + '_home'),
            on=['device_id'],
            how='inner'
        )
        return sdf

    @classmethod
    def set_dist_map(cls, distance):
        adm2_dist = pd.read_csv('../adm2_dist_nogeom.csv', index_col='L2_CODE')
        adm2_dist.drop(['Unnamed: 0'], axis=1, inplace=True)

        adm2_dist.columns = list(adm2_dist.columns[:8]) + list(adm2_dist.columns[8:].astype(float).astype(int))
        adm2_dist.iloc[:, 8:] = adm2_dist.iloc[:, 8:].astype(float) > distance

        cls.adm2_dist_dict = adm2_dist.iloc[:, 8:].to_dict()

    @staticmethod
    def count_outside_home(x, y, end_at_home):
        hah_score = 0

        x = [i for _, i in sorted(zip(y, x))]
        y.sort()

        outside_total = 0.0
        outside_initial = None

        for i in range(len(y) - 1):
            if outside_initial is None and x[i]:
                outside_initial = i

            if x[i] != x[i + 1]:

                outside_total += (y[i + 1] - y[i]).days / 2

                if not x[i + 1]:

                    outside_total += (y[i] - y[outside_initial]).days
                    outside_initial = None

                    if hah_score == 1:
                        hah_score += 1

                elif not hah_score:
                    hah_score += 1

        if hah_score != 2:
            return 1000.0

        if x[-1]:
            if end_at_home:
                return 1000.0

            if outside_initial is not None:
                outside_total += (y[-1] - y[outside_initial]).days

        return outside_total

    @staticmethod
    def count_outside_home_at_a_stretch(x, y, end_at_home):

        outside_home_days = 0

        x = [i for _, i in sorted(zip(y, x))]
        y.sort()

        outside_total = 0.0
        outside_total_t = 0.0
        outside_initial = None

        for i in range(len(y) - 1):
            if x[i]: outside_home_days += 1
            if outside_initial is None and x[i]:
                outside_initial = i
                if i > 0:
                    outside_total_t = (y[i] - y[i - 1]).days / 2

            if x[i] != x[i + 1]:
                if (outside_initial is not None) and (not x[i + 1]):
                    temp = outside_total_t + (y[i] - y[outside_initial]).days + (y[i + 1] - y[i]).days / 2
                    if temp > outside_total:
                        outside_total = temp
                    outside_total_t = 0
                    outside_initial = None

        if x[-1]:

            if end_at_home:
                return 0.0
            if outside_initial is not None:
                temp = outside_total_t + (y[-1] - y[outside_initial]).days
                if temp > outside_total:
                    outside_total = temp
            else:
                temp = (y[-1] - y[-2]).days / 2
                if temp > outside_total:
                    outside_total = temp
        if outside_home_days < 3:
            return 0.0

        return outside_total

    def _export(self, export_location):
        # this is where things need to be changed
        # add another column, is_outside_home which gives if someone is outside his/her home on some date
        # sjr_2019_1 = sjr_2019_1.withColumn("is_outside_home", col("admin2_home") != col("L2_CODE"))

        if self.distance_filter:
            dictionary = GenericShortTermMigration.adm2_dist_dict.copy()

        def is_outside_home(x, y):
            x = int(x)
            y = int(y)
            if x not in dictionary or y not in dictionary:
                return True

            return dictionary[x][y]

        is_outside_home_udf = F.udf(is_outside_home, T.BooleanType())

        if self.distance_filter:
            self.sjr = self.sjr.withColumn(
                "is_outside_home", is_outside_home_udf(
                    col(self.home_admin + '_home'),
                    col(self.home_admin),
                    lit(self.distance_filter),
                )
            )
        else:
            self.sjr = self.sjr.withColumn("is_outside_home", col(self.home_admin + '_home') != col(self.home_admin))

        if self.at_a_stretch:
            count_outside_home_udf = F.udf(self.count_outside_home_at_a_stretch, T.FloatType())
        else:
            count_outside_home_udf = F.udf(self.count_outside_home, T.FloatType())

        # group by device_id and aggregate using the custom user defined aggregate function
        device_id_aggregates = self.sjr.groupby('device_id', self.aggr_admin + '_home').agg(
            count_outside_home_udf(
                F.collect_list('is_outside_home'),
                F.collect_list('date'),
                lit(self.end_at_home),
            ).alias(
                'days_outside_home'
            )
        )

        # find number of STMs and total population per admin4
        stms = device_id_aggregates.filter(
            (col('days_outside_home') >= self.stm_range[0]) & (col('days_outside_home') <= self.stm_range[1])).groupby(
            self.aggr_admin + '_home').count().withColumnRenamed('count', 'no_of_stms')
        #         total_samples = device_id_aggregates.groupby(
        #             self.aggr_admin + '_home'
        #         ).count().withColumnRenamed('count', 'total_samples')

        total_samples = device_id_aggregates.join(self.ids, on="device_id", how="inner").groupby(
            self.aggr_admin + '_home'
        ).count().withColumnRenamed('count', 'total_samples')

        # bring the mobility population and total population together
        admin_stats = total_samples.join(stms, self.aggr_admin + '_home')

        # save the mobility stats for the month in a CSV file
        admin_stats.coalesce(1).write.format(
            "com.databricks.spark.csv"
        ).mode('overwrite').option("header", "true").save(
            export_location
        )

    def preprocess(self, month):

        # convert integer value of month to string format padded by one zero if necessary
        self.month = str(month).zfill(2)

        start_date = "2020/{}/01".format(month)

        if not self.fixed_home_period:
            self.set_device_id_homes(((datetime.strptime(start_date, '%Y/%m/%d') - relativedelta.relativedelta(
                months=12)).strftime('%Y/%m/%d'), None), 12)

        start_date_with_padding = (
                datetime.strptime(start_date, "%Y/%m/%d") - relativedelta.relativedelta(days=self.left_padding_days)
        ).strftime("%Y/%m/%d")

        end_date_with_padding = (
                datetime.strptime(start_date, "%Y/%m/%d") + relativedelta.relativedelta(
            months=self.timeframe_months) + relativedelta.relativedelta(days=self.right_padding_days)
        ).strftime("%Y/%m/%d")

        # extract data of no_of_months number of months starting from start_month
        self.sjr = self.extract_data_by_ends(self.sjr_master, start_date_with_padding, end_date_with_padding)

        # convert UNIX timestamp to date format
        self.sjr = self.sjr.withColumn('date', from_unixtime('timestamp').cast(T.DateType()))

        self.sjr = self.date_device_mode(self.sjr)

        self.sjr = self.sjr.join(self.device_id_homes, on=['device_id'], how='inner')

        # quality control filter
        self.sjr = self.quality_control_filter(self.sjr, 3)

        self.ids = self.extract_data(self.sjr, start_date, 1, by='date').select('device_id').distinct()

    def export_for_a_month(self, month, export_location):
        """
        Parameters:
          month: integer value of the month (example: 1, 2, ..., 12)
        """

        # preprocess the data to extract data of our time interval and apply quality control filter
        self.preprocess(month)
        self._export(export_location)



from pyspark.sql import SparkSession
from mobility.odm import OriginDestinationMigration


spark = SparkSession.builder.master("local").appName("mobility").getOrCreate()
od_matrix = OriginDestinationMigration(spark, 'L4_CODE', 'test_1.csv', 'haha', 1)
print(od_matrix.get_od_matrix("2019/01/01", "2020/10/10"))



import pandas as pd

from pyspark.sql import SparkSession

from .stm import GenericShortTermMigration
from aggregation import CustomAggregation


def main(start_month, end_month, **kwargs):

    master_df = None

    spark = SparkSession.builder.master("local").appName("mobility").getOrCreate()

    stm = GenericShortTermMigration(
        spark_session=spark,
        raw_file_location='/mnt/CUBEIQ/esapv/India/delta_unacast_v4',
        raw_file_format='delta',
        aggr_admin='adm4_code',
        home_admin='adm4_code',
        stm_range=[15, float('inf')],
        left_padding_days=7,
        right_padding_days=7,
        fixed_home_period=False
    )

    aggregation = CustomAggregation(
        aggr_admin='adm4_code',
        misc_files_dir='/dbfs/FileStore/tables/',
        province_column='adm1_name',
    )

    for month in range(start_month, end_month + 1):

        spark_export_location = \
            '/mnt/CUBEIQ/esapv/India/Unacast/Short Term Migration/One Month Definition/Test/2020_{}_7P/'.format(
            str(month).zfill(2))

        stm.export_for_a_month(month, export_location=spark_export_location)

        df = aggregation.aggregate('/dbfs/' + spark_export_location)

        df['Month'] = '2020/' + str(month).zfill(2)

        if master_df is None:
            master_df = df
        else:
            master_df = pd.concat([master_df, df], ignore_index=True)
        print(master_df.shape)

    master_df.to_csv(
        '/dbfs/mnt/CUBEIQ/esapv/India/Unacast/Short Term Migration/One Month Definition/Test/province_wise_stm_7_days_padding_variable.csv')


if __name__ == "__main__":
    main(1, 2)




import time
from itertools import chain
from datetime import datetime
from dateutil import relativedelta
from pyspark.sql import functions as F
from pyspark.sql.window import Window
from pyspark.ml.feature import Bucketizer
from pyspark.sql.types import IntegerType


class CommuterAggregation:

    def __init__(self, adm_col, delta_loc, record_filter=5, spark=None):
        self.adm_col = adm_col
        self.sjr = spark.read.format('delta').option("header", "true").load(delta_loc)

    @staticmethod
    def extract(sdf, from_date, to_date):
        from_t = time.mktime(datetime.strptime(from_date, "%H/%M").timetuple())
        to_t = time.mktime(datetime.strptime(to_date, "%H/%M").timetuple())

        timestamps_t = (from_t, to_t)
        mobility_t = sdf.where(F.col('timestamp').between(*timestamps_t))

        return mobility_t

    @staticmethod
    def get_hour_appended_data(sdf, buckets):
        sdf_with_date = sdf.withColumn("date", F.from_unixtime("timestamp", format="yyyy-MM-dd"))
        sdf_with_hour = sdf_with_date.withColumn("hour", F.from_unixtime("timestamp", format="HH").cast(IntegerType()))

        bucketizer = Bucketizer(splits=buckets, inputCol="hour", outputCol="bucket")
        hour_data = bucketizer.setHandleInvalid("keep").transform(sdf_with_hour)

        dict_items = [(0, f"{buckets[0]}-{buckets[1] + 1}")] + [(i, f"{buckets[i] + 1}-{buckets[i + 1] + 1}") for i in
                                                                range(1, len(buckets) - 1)]
        mapping = F.create_map([F.lit(x) for x in chain(*dict_items)])
        hour_data = hour_data.withColumn('hour_range', mapping[hour_data['bucket']])

        return hour_data

    def get_record_aggregates(self, periods):
        periods = [periods[0]] + [periods[i] - 1 for i in range(1, len(periods))]

        sjr_with_date = self.sjr.withColumn("date", F.from_unixtime("timestamp", format="yyyy-MM-dd"))

        hour_data = self.get_hour_appended_data(self.sjr, periods)
        hour_data = hour_data.groupby(self.adm_col, "hour_range").count()

        start = datetime.strptime(sjr_with_date.agg({'date': 'min'}).toPandas()['min(date)'][0], '%Y-%m-%d')
        end = datetime.strptime(sjr_with_date.agg({'date': 'max'}).toPandas()['max(date)'][0], '%Y-%m-%d')

        aggregates = hour_data.withColumn('daily_records', F.col('count') / (end - start).days).drop('count')

        return aggregates

    def get_device_aggregates(self, periods):
        periods = [periods[0]] + [periods[i] - 1 for i in range(1, len(periods))]

        hour_data = self.get_hour_appended_data(self.sjr, periods)
        hour_data = hour_data.dropDuplicates([self.adm_col, 'date', 'hour_range', 'device_id'])
        hour_data = hour_data.groupby(self.adm_col, "date", "hour_range").count()

        aggregates = hour_data.groupby(self.adm_col, 'hour_range').agg({'count': 'mean'}).withColumnRenamed(
            "avg(count)", "daily_devices")

        return aggregates


from collections import defaultdict
import pandas as pd


def get_net_migration_from_od_matrix(od_matrix_file_path, origin_col_name, dest_col_name, count_col_name, adm_name):
    df = pd.read_csv(od_matrix_file_path)

    outflow = defaultdict(int, dict(df.groupby(origin_col_name)[count_col_name].sum()))
    inflow = defaultdict(int, dict(df.groupby(dest_col_name)[count_col_name].sum()))

    admins = set(outflow.keys()).union(set(inflow.keys()))

    net_mvmt = {}

    for admin in admins:
        net_mvmt[admin] = inflow[admin] - outflow[admin]

    result = pd.DataFrame(net_mvmt.items(), columns=[adm_name, 'net_mvmt'])
    result['total'] = result[adm_name].map(outflow)

    return result


def get_avg_dist_per_dest(od_matrix_file_path, dist_matrix_file_path, adm_code_col_name, origin_col_name='origin',
                          destination_col_name='dest', count_col_name='count(1)'):
    od_matrix = pd.read_csv(od_matrix_file_path)
    dist_matrix = pd.read_csv(dist_matrix_file_path, index_col=adm_code_col_name)

    od_matrix['dist'] = od_matrix.apply(
        lambda x: dist_matrix.loc[x[origin_col_name], str(x[destination_col_name]).zfill(3)], axis=1)

    groups = od_matrix.groupby(destination_col_name)
    od_matrix['temp'] = od_matrix['dist'] * od_matrix[count_col_name] / groups[count_col_name].transform('sum')

    return dict(groups['temp'].sum())



import re
import time
from datetime import datetime
from dateutil import relativedelta
from pyspark.sql import functions as F
from pyspark.sql.window import Window


class OriginDestinationMigration:

    def __init__(self, adm_col, delta_loc, out_loc, record_filter=5, spark=None):
        self.spark = spark
        self.adm_col = adm_col
        self.delta_loc = delta_loc
        self.out_loc = out_loc
        self.record_filter = record_filter

        self.sjr = spark.read.format('delta').option("header", "true").load(delta_loc)

    @staticmethod
    def quality_control_filter(sdf, records):
        # filter out the IDs which have less than "records" number of records
        d = sdf.groupby('device_id').count().where(F.col('count') >= records)
        return sdf.join(d, on=["device_id"], how="inner")

    @staticmethod
    def extract(sdf, from_date, to_date):
        from_t = time.mktime(datetime.strptime(from_date, "%Y/%m/%d").timetuple())
        to_t = time.mktime(datetime.strptime(to_date, "%Y/%m/%d").timetuple())

        timestamps_t = (from_t, to_t)
        mobility_t = sdf.where(F.col('timestamp').between(*timestamps_t))

        return mobility_t

    def _get_od_matrix(self, from_date, to_date):
        extracted_data = self.extract(self.sjr, from_date, to_date)

        quality_filtered_data = self.quality_control_filter(extracted_data, self.record_filter)

        w = Window.partitionBy('device_id')
        max_table = quality_filtered_data.withColumn('maxtimestamp', F.max('timestamp').over(w)).where(
            F.col('timestamp') == F.col('maxtimestamp')).dropDuplicates()

        min_table = quality_filtered_data.withColumn('mintimestamp', F.min('timestamp').over(w)).where(
            F.col('timestamp') == F.col('mintimestamp')).dropDuplicates()

        # SQL operations
        max_table.createOrReplaceTempView('maxtable')
        min_table.createOrReplaceTempView('mintable')

        merged_table = self.spark.sql(
            f"""
                SELECT mxt.device_id, mxt.{self.adm_col} as dest, mnt.{self.adm_col} as origin
                FROM maxtable AS mxt JOIN mintable AS mnt
                ON mxt.device_id = mnt.device_id
            """
        )

        merged_table.createOrReplaceTempView('mergedtable')
        grouped_data = merged_table.groupBy('origin', 'dest')
        result = (grouped_data.agg({"*": "count"}))

        return result

    def get_od_matrix(self, from_date, to_date):
        return self._get_od_matrix(from_date, to_date)

    def get_net_migration(self, from_date, to_date):
        od_matrix = self.get_od_matrix(from_date, to_date)

        initial_samples = od_matrix.groupby('origin').agg({'count(1)': 'sum'}).withColumnRenamed('sum(count(1))',
                                                                                                 'origin_total')
        final_samples = od_matrix.groupby('dest').agg({'count(1)': 'sum'}).withColumnRenamed('sum(count(1))',
                                                                                             'dest_total')

        joined_samples = initial_samples.join(final_samples, initial_samples.origin == final_samples.dest,
                                              how='outer').fillna(0, ['origin_total', 'dest_total'])
        joined_samples = joined_samples.withColumn(self.adm_col, F.coalesce(*[i for i in ['origin', 'dest']])).drop(
            'origin', 'dest')
        joined_samples = joined_samples.withColumn('net_movement', F.col('dest_total') - F.col('origin_total'))
        joined_samples = joined_samples.withColumn('net_movement_pct',
                                                   F.col('net_movement') / F.col('origin_total')).drop('origin_total',
                                                                                                       'dest_total')

        return joined_samples

    def get_od_matrices(self, from_date, to_date, period='weekly'):
        if period not in set(['daily', 'weekly', 'fortnightly', 'monthly']):
            raise Exception("Period can only be either of 'daily', 'weekly', 'fortnightly' and 'monthly'!")

        pattern = re.compile("([12]\d{3}\/(0[1-9]|1[0-2])\/(0[1-9]|[12]\d|3[01]))$")

        if not pattern.match(from_date):
            raise Exception("from_data can only take date of the format YYYY/MM/DD!")

        if not pattern.match(to_date):
            raise Exception("to_date can only take date of the format YYYY/MM/DD!")

        period_map = {
            'daily': 1,
            'weekly': 7,
            'fortnightly': 15
        }

        if period in period_map:
            delta_time = relativedelta.relativedelta(days=period_map[period])
        else:
            delta_time = relativedelta.relativedelta(months=1)

        if datetime.strptime(from_date, '%Y/%m/%d') + delta_time > datetime.strptime(to_date, '%Y/%m/%d'):
            raise Exception(
                "Can not find even a single {} period in the date range {} to {}".format(period, from_date, to_date)
            )

        running_date = datetime.strptime(from_date, '%Y/%m/%d') + delta_time
        end_date = datetime.strptime(to_date, '%Y/%m/%d')

        while running_date <= end_date:
            od_matrix = od_migration.get_od_matrix(from_date, running_date.strftime("%Y/%m/%d"))

            # write the OD matrix in CSV format
            od_matrix.coalesce(1).write.format(
                "com.databricks.spark.csv"
            ).mode(
                'overwrite'
            ).option(
                "header", "true"
            ).save(
                self.out_loc + '/' + period.title() + '/' + running_date.date().__str__().replace('-', '_')
            )

            from_date = running_date.strftime("%Y/%m/%d")
            running_date = running_date + delta_time


class EventOriginDestinationMigration:

    def __init__(self, adm_col, delta_loc, out_loc, left_padding_days, right_padding_days, spark=None, record_filter=0):
        self.spark = spark
        self.adm_col = adm_col
        self.delta_loc = delta_loc
        self.out_loc = out_loc
        self.left_padding_days = left_padding_days
        self.right_padding_days = right_padding_days

        self.sjr = spark.read.format('delta').option("header", "true").load(delta_loc)
        self.record_filter = record_filter

    @staticmethod
    @F.udf
    def mode(x):
        from collections import Counter
        return Counter(x).most_common(1)[0][0]

    @staticmethod
    def quality_control_filter(sdf, records):
        # filter out the IDs which have less than "records" number of records
        d = sdf.groupby('device_id').count().where(F.col('count') >= records)
        return sdf.join(d, on=["device_id"], how="inner")

    @staticmethod
    def extract(sdf, from_date, to_date):
        from_t = time.mktime(datetime.strptime(from_date, "%Y/%m/%d").timetuple())
        to_t = time.mktime(datetime.strptime(to_date, "%Y/%m/%d").timetuple())

        timestamps_t = (from_t, to_t)
        return sdf.where(F.col('timestamp').between(*timestamps_t))

    def get_date_ends(self, event_date):
        from_date = (
                datetime.strptime(event_date, "%Y/%m/%d") - relativedelta.relativedelta(days=self.left_padding_days)
        ).strftime("%Y/%m/%d")

        to_date = (
                datetime.strptime(event_date, "%Y/%m/%d") + relativedelta.relativedelta(days=self.right_padding_days)
        ).strftime("%Y/%m/%d")
        return (from_date, to_date)

    def _get_od_matrix(self, from_date, to_date):
        extracted_data = self.extract(self.sjr, from_date, to_date)

        quality_filtered_data = self.quality_control_filter(extracted_data, self.record_filter)

        w = Window.partitionBy('device_id')
        max_table = quality_filtered_data.withColumn('maxtimestamp', F.max('timestamp').over(w)).where(
            F.col('timestamp') == F.col('maxtimestamp')).dropDuplicates()

        min_table = quality_filtered_data.withColumn('mintimestamp', F.min('timestamp').over(w)).where(
            F.col('timestamp') == F.col('mintimestamp')).dropDuplicates()

        # SQL operations
        max_table.createOrReplaceTempView('maxtable')
        min_table.createOrReplaceTempView('mintable')

        merged_table = self.spark.sql(
            f"""
                SELECT mxt.device_id, mxt.{self.adm_col} as dest, mnt.{self.adm_col} as origin
                FROM maxtable AS mxt JOIN mintable AS mnt
                ON mxt.device_id = mnt.device_id
            """
        )

        merged_table.createOrReplaceTempView('mergedtable')
        grouped_data = merged_table.groupBy('origin', 'dest')
        result = (grouped_data.agg({"*": "count"}))

        return result

    def get_od_matrix(self, event_date):
        from_date, to_date = self.get_date_ends(event_date)

        return self._get_od_matrix(from_date, to_date)

    def get_net_migration(self, event_date):
        od_matrix = self.get_od_matrix(event_date)

        initial_samples = od_matrix.groupby('origin').agg({'count(1)': 'sum'}).withColumnRenamed('sum(count(1))',
                                                                                                 'origin_total')
        final_samples = od_matrix.groupby('dest').agg({'count(1)': 'sum'}).withColumnRenamed('sum(count(1))',
                                                                                             'dest_total')

        joined_samples = initial_samples.join(final_samples, initial_samples.origin == final_samples.dest,
                                              how='outer').fillna(0, ['origin_total', 'dest_total'])
        joined_samples = joined_samples.withColumn(self.adm_col, F.coalesce(*[i for i in ['origin', 'dest']])).drop(
            'origin', 'dest')
        joined_samples = joined_samples.withColumn('net_movement', F.col('dest_total') - F.col('origin_total'))
        joined_samples = joined_samples.withColumn('net_movement_pct',
                                                   F.col('net_movement') / F.col('origin_total')).drop('origin_total',
                                                                                                       'dest_total')

        return joined_samples
